# -*- coding: utf-8 -*-
"""Fundamental models for rpi2caster, not depending on database.
Classes-only module. All constants come from definitions."""

from collections import defaultdict, OrderedDict
from contextlib import suppress
from itertools import chain
from math import ceil
from . import data, definitions as d

# Style and variant definitions for lookup
STYLES = {s.short: s for s in d.STYLES}
STYLES.update({s: s for s in d.STYLES})
STYLES.update({s.name: s for s in d.STYLES})


class Stats:
    """Casting statistics gathering and displaying functions"""
    def __init__(self):
        self._ribbon = defaultdict(int)
        self._run, self._session = defaultdict(int), defaultdict(int)
        self._current = defaultdict(lambda: defaultdict(str))
        self._previous = defaultdict(lambda: defaultdict(str))

    def update(self, **kwargs):
        """Update parameters"""
        def session_line_skip(lines):
            """skip lines for every run in the session"""
            self._session['lines_skipped'] = lines

        def run_line_skip(lines):
            """skip lines for a single run"""
            self._run['lines_skipped'] = lines

        def add_runs(delta):
            """change a number of runs in casting session"""
            old_value = self._session['runs']
            runs = old_value + delta
            self._session['runs'] = runs
            self._session['codes'] = runs * self._ribbon['codes']
            self._session['chars'] = runs * self._ribbon['chars']
            self._session['lines'] = runs * self._ribbon['lines']

        def end_run(success):
            """Updates the info about the runs"""
            if success:
                # update the counters
                self._session['runs_done'] += 1
                self._session['current_run'] += 1
            else:
                # reset last run stats - don't update the counters
                self._session['failed_runs'] += 1
                for par in ('current_line', 'current_code', 'current_char'):
                    self._session[par] -= self._run[par]

        parameters = dict(ribbon=self.update_session_stats,
                          queue=self.update_run_stats,
                          record=self.update_record_stats,
                          session_line_skip=session_line_skip,
                          run_line_skip=run_line_skip,
                          runs=add_runs,
                          casting_success=end_run)
        for argument, value in kwargs.items():
            routine = parameters.get(argument)
            if not routine:
                continue
            return routine(value)

    @property
    def runs(self):
        """Gets the runs (repetitions) number"""
        return self._session['runs']

    def get_ribbon_lines(self):
        """Gets number of lines per ribbon"""
        return self._ribbon['lines']

    def get_run_lines_skipped(self):
        """Get a number of lines to skip - decide whether to use
        the run or ribbon skipped lines"""
        return self._run['lines_skipped'] or self._session['lines_skipped']

    @property
    def ribbon_parameters(self):
        """Gets ribbon parameters"""
        parameters = OrderedDict({'': 'Ribbon parameters'})
        parameters['Combinations in ribbon'] = self._ribbon['codes']
        parameters['Characters incl. spaces'] = self._ribbon['chars']
        parameters['Lines to cast'] = self._ribbon['lines']
        return parameters

    @property
    def session_parameters(self):
        """Displays the ribbon data"""
        # display this only for multi-run sessions
        if self._session['runs'] < 2:
            return {}
        parameters = OrderedDict({'': 'Casting session parameters'})
        parameters['Casting runs'] = self._session['runs']
        parameters['All codes'] = self._session['codes']
        parameters['All chars'] = self._session['chars']
        parameters['All lines'] = self._session['lines']
        return parameters

    @property
    def code_parameters(self):
        """Displays info about the current combination"""
        def build_info(source, name):
            """Builds data to display based on the given parameter"""
            current_name = 'current_{}'.format(name)
            current_value = source[current_name]
            total_name = '{}s'.format(name)
            total_value = source[total_name]
            if not total_value:
                return '{current}'.format(current=current_value)
            done = max(0, current_value - 1)
            remaining = total_value - current_value
            percent = done / total_value * 100
            desc = ('{current} of {total} [{percent:.2f}%], '
                    '{done} done, {left} left')
            return desc.format(current=current_value, total=total_value,
                               percent=percent, done=done, left=remaining)

        record = self._current['record']
        runs_number = self._session['runs']
        # display comment as a header
        info = OrderedDict({'': record.comment})
        # display codes
        info['Raw signals from ribbon'] = record.signals
        if record.is_newline:
            info['Starting a new line'] = self._run['current_line']
        if runs_number > 1:
            info['Code / run'] = build_info(self._run, 'code')
            info['Char / run'] = build_info(self._run, 'char')
            info['Line / run'] = build_info(self._run, 'line')
            info['Run / job'] = build_info(self._session, 'run')
            info['Failed runs'] = self._session['failed_runs']
        info['Code / job'] = build_info(self._session, 'code')
        info['Char / job'] = build_info(self._session, 'char')
        info['Line / job'] = build_info(self._session, 'line')
        return info

    def update_session_stats(self, ribbon):
        """Parses the ribbon, counts combinations, lines and characters"""
        # first, reset counters
        self._session, self._ribbon = defaultdict(int), defaultdict(int)
        self._current = defaultdict(lambda: defaultdict(str))
        self._previous = defaultdict(lambda: defaultdict(str))
        # parse the ribbon to read its stats
        records = [record for record in ribbon if record.has_signals]
        # number of codes is just a number of valid combinations
        num_codes = len(records)
        # ribbon starts and ends with newline, so lines = newline codes - 1
        num_lines = sum(1 for record in records if record.is_newline) - 1
        # characters: if combination has no justification codes
        num_chars = sum(1 for record in records if record.is_char)
        # Set ribbon and session counters
        self._ribbon.update(codes=num_codes, chars=num_chars,
                            lines=max(0, num_lines))
        self._session.update(lines_skipped=0, current_run=1,
                             current_line=0, current_code=0, current_char=0)

    def update_run_stats(self, queue):
        """Parses the ribbon, counts combinations, lines and characters"""
        records = [record for record in queue if record.has_signals]
        # all codes is a number of valid signals
        num_codes = len(records)
        # clear and start with -1 line for the initial galley trip
        n_lines = sum(1 for record in records if record.is_newline) - 1
        num_lines = max(n_lines, 0)
        # determine number of characters
        num_chars = sum(1 for record in records if record.is_char)
        # update run statistics
        self._run.update(lines_skipped=0,
                         current_line=0, current_code=0, current_char=0,
                         lines=num_lines, codes=num_codes, chars=num_chars)

    def update_record_stats(self, record):
        """Updates the stats based on parsed signals."""
        # Update previous stats and current record
        self._previous, self._current['record'] = self._current.copy(), record
        # advance or set the run and session code counters
        self._run['current_code'] += 1
        self._session['current_code'] += 1
        # detect 0075+0005(+pos_0005)->0075(+pos_0075) = double justification
        # this means starting new line
        with suppress(AttributeError):
            was_newline = self._previous['record'].is_newline
            if was_newline and record.is_pump_start:
                self._run['current_line'] += 1
                self._session['current_line'] += 1
        # advance or set the character counters
        self._run['current_char'] += 1 * record.is_char
        self._session['current_char'] += 1 * record.is_char

    def get_runs_left(self):
        """Gets the runs left number"""
        diff = self._session['runs'] - self._session['runs_done']
        return max(0, diff)

    def get_lines_done(self):
        """Gets the current run line number"""
        return max(0, self._run['current_line'] - 1)


class Matrix:
    """A class for single matrices - all matrix data"""
    __slots__ = ('char', '_units', 'row', 'column', '_styles', 'diecase',
                 'position', 'size')

    def __init__(self, char='', styles='*', code='', units=0, diecase=None):
        self.diecase = diecase
        self.char = char
        self.size = d.MatrixSize(horizontal=1, vertical=1)
        self.position, self.code = d.Coordinates(None, None), code
        self._units = units
        self._styles = Styles(styles).string

    def __len__(self):
        return 0 if self.isspace() else len(self.char)

    def __bool__(self):
        return True

    def __repr__(self):
        return ('<Matrix: {m.position.column}{m.position.row} '
                '({m.char} {styles}, {m.units} units wide)>'
                .format(m=self, styles=self.styles.names))

    def __str__(self):
        if self.ishighspace():
            tmpl = 'high space at {pos.column}{pos.row}'
        elif self.islowspace():
            tmpl = 'low space at {pos.column}{pos.row}'
        else:
            tmpl = ('{m.char} ({m.styles.names}) at {pos.column}{pos.row}, '
                    '{m.units} wide')
        return tmpl.format(m=self, pos=self.position)

    @property
    def comment(self):
        """Generate a comment based on matrix character and style."""
        desc = ('low space {mat.ems:.2f} ems' if self.islowspace()
                else 'high space {mat.ems:.2f} ems' if self.ishighspace()
                else '{mat.char} ({mat.styles.names})' if self.char
                else 'unknown character')
        return ''.join((desc, ' at {mat.code}')).format(mat=self)

    @property
    def units(self):
        """Gets the specific or default number of units"""
        # If units are assigned to the matrix, return the value
        # Otherwise, wedge default
        if self.isspace() or not self.char:
            return self.get_units_from_row()
        else:
            return (self._units or self.get_units_from_arrangement() or
                    self.get_units_from_row())

    @units.setter
    def units(self, units):
        """Sets the unit width value. Do nothing with non-int values"""
        with suppress(TypeError):
            self._units = int(units)

    @property
    def ems(self):
        """Get the ems for matrix; 1em = 18 units"""
        return round(self.units / 18, 2)

    @property
    def code(self):
        """Gets the matrix code"""
        try:
            return '{pos.column}{pos.row}'.format(pos=self.position)
        except AttributeError:
            return ''

    @code.setter
    def code(self, codes):
        """Sets the coordinates for the matrix"""
        def row_generator():
            """Generate matching rows, removing them from input sequence"""
            # first is None as the sequence generated will be reversed
            # and None is to be used as a last resort if no row is found
            yield None
            nonlocal sigs
            for number in range(16, 0, -1):
                string = str(number)
                if string in sigs:
                    sigs = sigs.replace(string, '')
                    yield number

        def column_generator():
            """Generate column numbers"""
            nonlocal sigs
            for column in ['NI', 'NL', *'ABCDEFGHIJKLMNO']:
                if column in sigs:
                    sigs = sigs.replace(column, '')
                    yield column
            yield None

        # needs to work with strings and iterables
        sigs = ''.join(str(l) for l in codes or '').upper()

        # get a first possible row (caveat: recognize double-digit numbers)
        all_rows = [x for x in row_generator()]
        rows = (x for x in reversed(all_rows))
        # get the first possible column -> NI, NL, A...O
        columns = column_generator()
        self.position = d.Coordinates(column=next(columns), row=next(rows))

    @property
    def styles(self):
        """Get the matrix's styles or an empty string if matrix has no char"""
        if self.isspace() or not self.char:
            # all styles no matter what
            return Styles('*')
        else:
            return Styles(self._styles)

    @styles.setter
    def styles(self, styles):
        """Sets the matrix's style string"""
        self._styles = Styles(styles).string

    def islowspace(self):
        """Checks whether this is a low space: char is present"""
        return self.char and self.char.isspace()

    def set_lowspace(self, value):
        """Update self.char with " " for low spaces"""
        if value:
            self.char = ' '

    def ishighspace(self):
        """Checks whether this is a high space: "_" is present"""
        # Any number of _ is present, no other characters
        return (self.char and '_' in self.char and not
                [x for x in self.char.strip() if x != '_'])

    def set_highspace(self, value):
        """Set the character to _"""
        if value:
            self.char = '_'

    def isspace(self):
        """Check if it's a low or high space"""
        return self.islowspace() or self.ishighspace()

    def get_units_from_row(self, wedge_used=None):
        """Gets a number of units for characters in the diecase row"""
        try:
            wedge = wedge_used or self.diecase.wedge
            return wedge.units[self.position.row]
        except (AttributeError, TypeError, IndexError):
            return 0

    def get_min_units(self):
        """Gets the minimum unit value for a given wedge, based on the
        matrix row and wedge unit value, for wedges at 1/1"""
        wedge = self.diecase.wedge
        limits = wedge.get_adjustment_limits(low_space=self.islowspace())
        # at least one unit
        return round(max(self.get_units_from_row() - limits.shrink, 1), 2)

    def get_max_units(self):
        """Gets the minimum unit value for a given wedge, based on the
        matrix row and wedge unit value, for wedges at 1/1"""
        wedge = self.diecase.wedge
        limits = wedge.get_adjustment_limits(low_space=self.islowspace())
        return round(self.get_units_from_row() + limits.stretch, 2)

    def get_units_from_arrangement(self):
        """Try getting the unit width value from diecase's unit arrangement,
        if that fails, return 0"""
        for style in self.styles:
            with suppress(ValueError, KeyError, AttributeError):
                arrangement = self.diecase.unit_arrangements.get(style)
                if not arrangement:
                    continue
                return arrangement[self.char]
        return 0

    def get_layout_record(self, pos=True):
        """Returns a record suitable for JSON-dumping and storing in DB."""
        code = '{p.column}{p.row}'.format(p=self.position) if pos else ''
        return d.MatrixRecord(self.char, self.styles.string, code, self.units)

    def get_ribbon_record(self, s_needle=False, comment=True):
        """Returns a signals record for adding to ribbon.

        s_needle : whether to add "S" signal or not,
        comment : False or None for no comment,
                  True to auto-generate the comment,
                  any other value to use it as a comment
        """
        _comment = ('' if not comment
                    else self.comment if comment is True
                    else comment)
        code = ('{pos.column}{es}{pos.row}'
                .format(pos=self.position, es='S' * s_needle))
        commented = '{:<20} // {}'
        return commented.format(code, _comment) if _comment else code


class Styles:
    """Styles collection grouping styles and allowing for edit.
    styles: a) string, b) iterable of Style namedtuples, c) Styles object
    default: single Style namedtuple, or style short
    mask: input same as styles; this limits the choice of styles present
          to what is contained in the mask.
    """
    definitions = d.STYLES
    __slots__ = ('_styles', 'default', 'mask')

    def __init__(self, styles='*', default=None, mask=None):
        # which style is default, in case nothing is given?
        self.default = STYLES.get(default, d.STYLES.roman)
        # get a limiting mask: iterable, styles, single style
        try:
            style_mask = {STYLES.get(s) for s in mask}
            style_mask.discard(None)
        except TypeError:
            style_mask = {STYLES.get(mask)}
            style_mask.discard(None)
        self.mask = tuple(style_mask or self.definitions)

        # get styles from source and filter only the non-None ones
        try:
            source = (set(self.definitions) if styles == '*'
                      else {STYLES.get(s) for s in styles})
        except TypeError:
            source = {STYLES.get(styles)}
        # allow only non-None items, and if nothing is found, use default
        source.discard(None)
        source = source or {self.default}

        self._styles = tuple(s for s in self.definitions
                             if s in source and s in self.mask)

    def __iter__(self):
        return (x for x in self._styles)

    def __bool__(self):
        return True if self._styles else False

    def __str__(self):
        return self.string

    def __repr__(self):
        return '<Styles: {}>'.format(self.string)

    def __add__(self, other):
        if not other:
            return self
        try:
            return Styles(self.string + other.string)
        except AttributeError:
            return Styles(self.string)

    def __radd__(self, other):
        if not other:
            return self
        try:
            return Styles(other.string + self.string)
        except AttributeError:
            return Styles(self.string)

    def __contains__(self, what):
        _style = STYLES.get(what)
        return _style in self._styles

    def __eq__(self, other):
        try:
            # two styles collections having common elements compare equal
            # should work with strings and Styles objects as well
            return (set(self).intersection(other) or
                    set(str(self)).intersection(str(other)))
        except (TypeError, ValueError, AttributeError):
            return False

    def __getitem__(self, index):
        return self._styles[index]

    @property
    def string(self):
        """Return the string of all style short names"""
        return ('*' if self.use_all
                else ''.join(style.short for style in self._styles))

    @property
    def names(self):
        """Get the long names of styles"""
        return ('all styles' if self.use_all
                else ', '.join(style.name for style in self._styles))

    @property
    def all_names(self):
        """Get all names including alternatives"""
        names = ('{} ({})'.format(styledef.name, styledef.alternatives)
                 if styledef.alternatives else styledef.name
                 for styledef in self._styles)
        return 'all styles' if self.use_all else ', '.join(names)

    @property
    def use_all(self):
        """Check if the collection has every style"""
        return set(self) == set(self.mask)

    @property
    def is_single(self):
        """Check if only one style is in collection"""
        return len(self._styles) == 1

    @property
    def is_default(self):
        """Check if this is a default style and nothing else"""
        return set(self._styles) == {self.default}

    @property
    def first(self):
        """Get the first style from the collection"""
        for style in self._styles:
            return style
        return self.default

    def items(self):
        """Get items - so that the object can be iterated like a dict"""
        return ((style.name, style) for style in self._styles)

    def ansi_format(self, formatted_string):
        """Use ANSI escape sequences to format a string"""
        params = (str(style.ansi) for style in self._styles)
        par_string = ';'.join(params)
        if self.use_all or not par_string:
            return formatted_string
        else:
            start, end = '\033[', '\033[0m'
            return ''.join([start, par_string, 'm', formatted_string, end])


class Measure:
    """Chooses and represents a line length, space width etc. measures
    which can be expressed in various typographic units (for definition,
    see TYPOGRAPHIC_UNITS)."""
    __slots__ = ('unit', 'points', 'set_width')

    def __init__(self, input_value, unit='pt', set_width=12.0):
        if not input_value:
            raise ValueError('Cannot parse measure: {}'.format(input_value))
        # Support parsing of the relative Monotype units, based on set width
        # We know the set width now, and need to update the unit definitions
        units = d.TYPOGRAPHIC_UNITS.copy()
        units.update(em=round(set_width, 2), en=round(set_width / 2, 2),
                     u=round(set_width / 18, 2))
        # Parse the value
        raw = str(input_value).strip()
        string = ''.join(x for x in raw if x.isalnum() or x in ',.')
        # Get the unit suffix (end on first match), otherwise keep default
        self.unit = unit
        for symbol in sorted(units, reverse=True):
            if string.endswith(symbol):
                string = string.replace(symbol, '')
                self.unit = symbol
                break
        # Filter the string to remove all letters, round the value to 2
        num_string = ''.join(x for x in string if x in '.,0123456789')
        value = round(float(num_string.replace(',', '.')), 2)
        self.set_width = set_width
        self.points = value * units.get(self.unit)

    def __str__(self):
        return '{}{}'.format(self.value, self.unit)

    @property
    def value(self):
        """Get value in units defined on init"""
        return self.get_value()

    def get_value(self, unit=None):
        """Get a value expressed in specified or default units"""
        # Get the coefficient for calculation
        factor = 1 / d.TYPOGRAPHIC_UNITS.get(unit or self.unit, 1)
        return round(self.points * factor, 2)

    @property
    def ems(self):
        """Gets the number of ems of specified set width"""
        return round(self.points / self.set_width, 2)

    @property
    def units(self):
        """Calculates the line length in wedge's units"""
        return round(self.ems * 18, 2)

    @property
    def inches(self):
        """An absolute width of this measure in inches"""
        return self.points / 72.0


class CharFreqs:
    """Read and calculate char frequencies, translate that to casting order"""
    scale, case_ratio = 1.0, 1.0

    def __init__(self, lang=None):
        language = str(lang).strip()
        self.lang = language
        self.freqs = data.LETTER_FREQUENCIES[language.lower()]

    def __repr__(self):
        return '<CharFreqs for {}>'.format(d.LANGS[self.lang])

    def __str__(self):
        return d.LANGS.get(self.lang, '')

    def __iter__(self):
        return (char for char in sorted(self.freqs))

    def get_type_bill(self):
        """Returns an iterator object of tuples: (char, qty)
        for each character."""
        def quantity(char, upper=False):
            """Calculate character quantity based on frequency"""
            ratio = self.case_ratio if upper else 1
            normalized_qty = self.freqs.get(char, 0) / self.freqs.get('a', 1)
            return max(ceil(normalized_qty * self.scale * ratio), 10)

        # Start with lowercase
        lower_bill = ((char, quantity(char)) for char in sorted(self.freqs))
        upper_bill = ((char.upper(), quantity(char, upper=True))
                      for char in sorted(self.freqs) if char.isalpha())
        return chain(lower_bill, upper_bill)

    def set_scales(self, scale=1.0, case_ratio=1.0):
        """Set the scale factor and uppercase-to-lowercase ratio"""
        self.scale, self.case_ratio = scale, case_ratio


class MatrixNotFound(Exception):
    """Cannot find matrix in diecase layout."""


class TypesettingError(Exception):
    """Various exceptions raised when typesetting"""



# -*- coding: utf-8 -*-
"""Casting utility: cast or punch ribbon, cast material for hand typesetting,
make a diecase proof, quickly compose and cast text.
"""

from collections import deque, OrderedDict
from functools import wraps

# common definitions
import librpi2caster
# QR code generating backend
try:
    import qrcode
except ImportError:
    qrcode = None

from .rpi2caster import UI, Abort, Finish, option, find_casters
from . import basic_models as bm, basic_controllers as bc, definitions as d
from . import main_controllers as mc
from .parsing import parse_record
from .typesetting import TypesettingContext


def cast_this(ribbon_source):
    """Get the ribbon from decorated routine and cast it"""
    @wraps(ribbon_source)
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        ribbon = ribbon_source(self, *args, **kwargs)
        if not ribbon:
            return
        if self.machine.is_casting():
            return self.cast_ribbon(ribbon)
        else:
            return self.punch_ribbon(ribbon)
    return wrapper


class Casting(TypesettingContext):
    """Casting:

    Methods related to operating the composition caster.
    Requires configured caster.

    All methods related to operating a composition caster are here:
    -casting composition and sorts, punching composition,
    -calibrating the caster,
    -testing the interface,
    -sending an arbitrary combination of signals,
    -casting spaces to heat up the mould."""
    machine = None

    def __init__(self, interface_id, operation_mode=None):
        self.choose_machine(interface_id, operation_mode)

    def choose_machine(self, interface_id=None, operation_mode=None):
        """Choose a machine from the available interfaces."""
        def make_menu_entry(number, caster, url, name):
            """build a menu entry"""
            if caster:
                if number:
                    nums.append(number)
                modes = ', '.join(caster.supported_operation_modes)
                row16_modes = ', '.join(caster.supported_row16_modes) or 'none'
                text = ('{} - modes: {} - row 16 addressing modes: {}'
                        .format(caster, modes, row16_modes))
            else:
                text = '[Unavailable] {}\n\t\t\t{}'.format(url, name)
            return option(key=number, seq=number, value=caster, text=text)

        casters = find_casters(operation_mode)
        # look a caster up by the index
        try:
            url, caster, name = casters[interface_id]
            if not caster:
                raise ValueError
            UI.display('Using caster {} at {}'.format(name, url),
                       min_verbosity=1)
        except (KeyError, IndexError, TypeError, ValueError):
            # choose caster from menu
            # nums stores menu entry numbers for useful casters
            nums = []
            menu_options = [make_menu_entry(number, caster, url, name)
                            for number, (url, caster, name) in casters.items()]
            while True:
                caster = UI.simple_menu('Choose the caster:', menu_options,
                                        default_key=nums[0] if nums else 0)
                if not caster:
                    UI.pause('Tried to use the unavailable caster.')
                else:
                    break
        self.machine = caster

    def punch_ribbon(self, ribbon):
        """Punch the ribbon from start to end"""
        self.machine.punch(ribbon)

    def cast_ribbon(self, input_iterable):
        """Main casting routine.

        First check if ribbon needs rewinding (when it starts with pump stop).

        Ask user about number of repetitions (useful for e.g. business cards
        or souvenir lines), number of initial lines skipped for all runs
        and the upcoming run only.

        Ask user whether to pre-heat the mould
        to stabilize the temperature and cast good quality type.

        Cast the ribbon single or multiple times, displaying the statistics
        about current record and casting progress.

        If casting multiple runs, repeat until all are done,
        offer adding some more.
        """
        def rewind_if_needed():
            """Decide whether to rewind the ribbon or not.
            If casting and stop comes first, rewind. Otherwise not."""
            nonlocal ribbon
            for record in ribbon:
                if record.is_pump_stop:
                    # rewind
                    ribbon = [x for x in reversed(ribbon)]
                    return
                elif record.is_pump_start:
                    # no need to rewind
                    return

        def skip_lines(source):
            """Skip a definite number of lines"""
            # Apply constraints: 0 <= lines_skipped < lines in ribbon
            lines_skipped = stats.get_run_lines_skipped()
            if lines_skipped:
                UI.display('Skipping {} lines'.format(lines_skipped))
            # Take away combinations until we skip the desired number of lines
            # BEWARE: ribbon starts with galley trip!
            # We must give it back after lines are taken away
            record = parse_record('')
            sequence = deque(source)
            while lines_skipped > 0:
                record = sequence.popleft()
                lines_skipped -= 1 * record.is_newline
            # give the last code back
            sequence.appendleft(record)
            return sequence

        def set_lines_skipped(run=True, session=True):
            """Set the number of lines skipped for run and session
            (in case of multi-session runs)."""
            # allow skipping only if ribbon is more than 2 lines long
            limit = max(0, stats.get_ribbon_lines() - 2)
            if not limit:
                return
            # how many can we skip?
            if run or session:
                UI.display('We can skip up to {} lines.'.format(limit))

            if run:
                # run lines skipping
                # how many lines were successfully cast?
                lines_ok = stats.get_lines_done()
                if lines_ok:
                    UI.display('{} lines were cast in the last run.'
                               .format(lines_ok))
                # Ask user how many to skip (default: all successfully cast)
                r_skip = UI.enter('How many lines to skip for THIS run?',
                                  default=lines_ok, minimum=0, maximum=limit)
                stats.update(run_line_skip=r_skip)

            if session:
                # Skip lines effective for ALL runs
                # session line skipping affects multi-run sessions only
                # don't do it for single-run sessions
                if stats.runs > 1:
                    s_skip = UI.enter('How many lines to skip for ALL runs?',
                                      default=0, minimum=0, maximum=limit)
                    stats.update(session_line_skip=s_skip)

        def cast_queue(queue):
            """Casts the sequence of codes in given sequence.
            This function is executed until the sequence is exhausted
            or casting is stopped by machine or user."""
            # in punching mode, lack of row will trigger signal 15,
            # lack of column will trigger signal O
            # in punching and testing mode, signal O or 15 will be present
            # in the output combination as O15
            for record in queue:
                # check if signal will be cast at all
                if not record.has_signals:
                    UI.display_header(record.comment)
                    continue
                # display some info and cast the signals
                stats.update(record=record)
                # cast it and get current machine status
                status = self.machine.cast_one(record)
                # prepare data for display
                casting_status = OrderedDict()
                casting_status['Signals sent'] = ' '.join(status['signals'])
                casting_status['Pump'] = 'ON' if status['pump'] else 'OFF'
                casting_status['Wedge 0005 at'] = status['wedge_0005']
                casting_status['Wedge 0075 at'] = status['wedge_0075']
                casting_status['Speed'] = status['speed']
                # display status
                UI.clear()
                UI.display_parameters(stats.code_parameters, casting_status)

        # Ribbon pre-processing and casting parameters setup
        ribbon = [parse_record(code) for code in input_iterable]
        rewind_if_needed()
        # initialize statistics
        stats = bm.Stats()
        stats.update(ribbon=ribbon)
        # display some info for the user
        UI.display_parameters(stats.ribbon_parameters)
        # set the number of casting runs
        stats.update(runs=UI.enter('How many times do you want to cast this?',
                                   default=1, minimum=0))
        # initial line skipping
        set_lines_skipped(run=True, session=True)
        UI.display_parameters(stats.session_parameters)
        # check if the row 16 addressing will be used;
        # connect with the interface
        row16_in_use = any(record.uses_row_16 for record in ribbon)
        self.machine.choose_row16_mode(row16_in_use)
        with self.machine:
            while stats.get_runs_left():
                # Prepare the ribbon ad hoc
                queue = skip_lines(ribbon)
                stats.update(queue=queue)
                # Cast the run and check if it was successful
                try:
                    cast_queue(queue)
                    stats.update(casting_success=True)
                    if not stats.get_runs_left():
                        # make sure the machine will check
                        # whether it's running next time
                        self.machine.working = False
                        # user might want to re-run this
                        prm = 'Casting successfully finished. Any more runs?'
                        stats.update(runs=UI.enter(prm, default=0, minimum=0))

                except librpi2caster.MachineStopped:
                    stats.update(casting_success=False)
                    # aborted - ask if user wants to continue
                    runs_left = stats.get_runs_left()
                    if runs_left:
                        UI.confirm('{} runs left, continue?'.format(runs_left),
                                   default=True, abort_answer=False)
                    else:
                        UI.confirm('Retry casting?', default=True,
                                   abort_answer=False)
                    # offer to skip lines for re-casting the failed run
                    skip_successful = stats.get_lines_done() >= 2
                    set_lines_skipped(run=skip_successful, session=False)
                    # restart the machine after emergency stop
                    self.machine.start()

    @cast_this
    @mc.temp_wedge
    @bc.temp_measure
    def cast_material(self):
        """Cast typesetting material: typecases, specified sorts, spaces"""
        def make_queue():
            """generate a sequence of items for casting"""
            while True:
                try:
                    UI.display('Specify characters to cast.\n'
                               'When done, press Enter to start casting.')
                    matrix = self.find_matrix(char='')
                    units = UI.enter('How many units?', datatype=int,
                                     default=round(self.get_units(matrix)))
                    qty = UI.enter('How many sorts?', default=10, minimum=0)
                    # ready to deliver
                    yield d.QueueItem(matrix, round(units, 2), qty)
                except (StopIteration, bm.MatrixNotFound):
                    break

        def make_ribbon(queue):
            """Take items from queue and adds them as long as there is
            space left in the galley. When space runs out, end a line.

            queue: [(code, quantity, units, pos_0075, pos_0005)]"""
            def new_mat():
                """matrix, width --> ribbon code, wedge positions"""
                # if the generator yields None first, try again
                queue_item = next(queue) or next(queue)
                sorts_left = queue_item.qty
                matrix, units = queue_item.matrix, queue_item.units
                # some info for the user
                UI.display('{} × {}, {} units {} set'
                           .format(sorts_left, matrix.code, units,
                                   self.wedge.set_width))
                # get code and wedge positions for the item
                positions = self.get_wedge_positions(matrix, units)
                record = matrix.get_ribbon_record(s_needle=positions != (3, 8))
                return record, units, positions, sorts_left

            def newline():
                """fill the line with quads, then spaces"""
                nonlocal units_left
                # single justification for the characters
                coarse, fine = wedges
                char_justification = ['NKS 0075 {}'.format(coarse),
                                      'NJS 0005 {}'.format(fine)]

                # add quads (one extra - last quad in the row)
                n_quads = 1 + int(units_left // self.quad.units)
                units_left %= self.quad.units
                try:
                    # variable space to adjust the width
                    mat = self.find_space(units=units_left)
                    coarse, fine = self.get_wedge_positions(mat, units_left)
                    use_s_needle = (coarse, fine) != (3, 8)
                    var_space = [mat.get_ribbon_record(s_needle=use_s_needle)]
                except bm.MatrixNotFound:
                    var_space = []
                    # no space to justify = just 0075+0005+NKJ
                    coarse, fine = '', ''
                # double justification sets the initial space width
                if coarse == fine:
                    # single code is enough
                    space_justification = ['NKJS 0005 0075 {}'.format(fine)]
                else:
                    space_justification = ['NKS 0075 {}'.format(coarse),
                                           'NKJS 0005 0075 {}'.format(fine)]

                units_left = self.measure.units - 2 * self.quad.units
                # single justification (for type), fillup spaces & quads,
                # double justification (for space), initial quad on new line
                return [*char_justification, *var_space, *[quad] * n_quads,
                        *space_justification, quad]

            def changeover():
                """use single-justification (0005+0075) to adjust wedges"""
                # add a quad between different series
                nonlocal units_left
                units_left -= self.quad.units * 2
                coarse, fine = wedges
                quads = [quad, quad]
                # single justification
                sjust = ([] if (coarse, fine) == (3, 8)
                         else ['NKS 0075 {}'.format(fine)] if fine == coarse
                         else ['NKS 0075 {}'.format(coarse),
                               'NJS 0005 {}'.format(fine)])
                return [*quads, *sjust]

            def add_code():
                """add codes to a ribbon, updating the number in the process"""
                nonlocal units_left, sorts_left
                # how many can we fit in the line? (until we've cast all)
                number = int(min(units_left // units, sorts_left))
                # update counters
                sorts_left -= number
                units_left -= number * units
                return [record] * number

            # em-quad for filling the line
            quad = self.quad.get_ribbon_record()
            # initialize the units
            units_left = self.measure.units - 2 * self.quad.units
            # get the first matrix
            # (if StopIteration is raised, no casting)
            record, units, wedges, sorts_left = new_mat()
            # first to set / last to cast last line out
            yield ['NJS 0005', 'NKJS 0005 0075', quad]
            # keep adding these characters
            while True:
                yield add_code()
                if sorts_left:
                    # still more to go...
                    yield newline()
                else:
                    # we're out of sorts... next character
                    try:
                        yield changeover()
                        record, units, wedges, sorts_left = new_mat()
                    except StopIteration:
                        # no more characters => fill the line and finish
                        # those are the first characters to cast
                        yield newline()
                        break
                    except (bm.TypesettingError, bm.MatrixNotFound) as exc:
                        UI.display('{}, omitting'.format(exc))
                        continue

        source = make_queue()
        ribbon = [code for chunk in make_ribbon(source) for code in chunk]
        return ribbon

    @cast_this
    @mc.temp_wedge
    def cast_qr_code(self):
        """Set up and cast a QR code which can be printed and then scanned
        with a mobile device."""
        def define_space(low):
            """find and set up a high or low space"""
            what = 'Low' if low else 'High'
            char = ' ' if low else '_'
            code = UI.enter('{} space matrix coordinates?'.format(what), '')
            space = bm.Matrix(char=char, code=code, diecase=self.diecase)
            wedges = self.get_wedge_positions(space, units)
            return space.get_ribbon_record(s_needle=wedges != (3, 8)), wedges

        def make_qr(data):
            """make a QR code matrix from data"""
            # QR rendering parameters
            border = UI.enter('QR code border width (squares)?', default=1,
                              minimum=1, maximum=10)
            ec_option = UI.enter('Error correction: 0 = lowest, 3 = highest?',
                                 default=1, minimum=0, maximum=3)
            # set up a QR code and generate a matrix
            modes = (qrcode.constants.ERROR_CORRECT_L,
                     qrcode.constants.ERROR_CORRECT_M,
                     qrcode.constants.ERROR_CORRECT_H,
                     qrcode.constants.ERROR_CORRECT_Q)
            engine = qrcode.QRCode(error_correction=modes[ec_option],
                                   border=border)
            engine.add_data(data)
            qr_matrix = engine.get_matrix()
            return qr_matrix

        def render(pattern):
            """translate a pattern into Monotype control codes,
            applying single justification if space widths differ,
            making spaces square in shape"""
            characters = {False: (low_space, ls_wedges),
                          True: (high_space, hs_wedges)}
            ribbon = ['NJS 0005', 'NKJS 0075 0005']
            for line in pattern:
                pairs = zip(line, [*line[1:], None])
                # newline (border is always low space)
                # double justification with low space wedges
                for current_item, next_item in pairs:
                    # add all spaces in a row
                    space, wedges = characters.get(current_item)
                    try:
                        _, next_wedges = characters.get(next_item)
                    except TypeError:
                        next_wedges = ls_wedges
                    ribbon.append(space)
                    if wedges != (3, 8) and wedges != next_wedges:
                        # set the wedges only if we need to
                        # use single justification in this case
                        ribbon.append('NKS 0075 {}'.format(wedges.pos_0075))
                        ribbon.append('NJS 0005 {}'.format(wedges.pos_0005))
                ribbon.append('NKS 0075 {}'.format(ls_wedges.pos_0075))
                ribbon.append('NKJS 0005 0075 {}'.format(ls_wedges.pos_0005))
            return ribbon

        # set the pixel size; smaler is preferred; depends on mould
        # (allow using different typesetting measures)
        px_size = bc.set_measure('6pt', what='pixel size (the same as mould)',
                                 set_width=self.wedge.set_width)
        units = px_size.units
        UI.display('The pixel is {} units {} set wide.'
                   .format(units, self.wedge.set_width))
        # determine the low and high space first
        try:
            low_space, ls_wedges = define_space(True)
            high_space, hs_wedges = define_space(False)
        except bm.TypesettingError as exc:
            UI.display(str(exc))
            UI.pause('Try again with a different pixel size or wedge')
            return self.cast_qr_code()
        # enter text and encode it
        text = UI.enter('Enter data to encode', '')
        qr_matrix = make_qr(text)
        # let the operator know how large the code is
        size = len(qr_matrix)
        prompt = ('The resulting QR code is {0} × {0} squares '
                  'or {1} × {1} inches.')
        UI.display(prompt.format(size, round(size * px_size.inches, 1)))
        UI.pause('Set your galley accordingly or abort.', allow_abort=True)
        # make a sequence of low and high spaces to cast
        return render(qr_matrix)

    @cast_this
    def cast_composition(self):
        """Casts or punches the ribbon contents if there are any"""
        if not self.ribbon.contents:
            raise Abort
        return self.ribbon.contents

    @cast_this
    @mc.temp_diecase
    @mc.temp_wedge
    def diecase_proof(self):
        """Tests the whole diecase, casting from each matrix.
        Casts spaces between characters to be sure that the resulting
        type will be of equal width."""
        def get_codes(matrix, wedges):
            """add codes with single justification if necessary"""
            combinations = []
            wedges_needed = wedges != (3, 8)
            code = matrix.get_ribbon_record(s_needle=wedges_needed)
            combinations.append(code)
            if wedges_needed:
                combinations.append('NKS 0075 {}'.format(wedges.pos_0075))
                combinations.append('NJS 0005 {}'.format(wedges.pos_0005))
            return combinations

        def add_matrix(matrix):
            """a matrix and a space"""
            # keep track of spaces too narrow to cast
            nonlocal leftover_units
            mat_units = self.get_units(matrix)
            space_units = 22 + leftover_units - mat_units
            leftover_units = 0
            try:
                mat_wedges = self.get_wedge_positions(matrix, mat_units)
            except bm.TypesettingError:
                # adjust char width as far as possible,
                # take away or add the rest to the space instead
                limits = self.wedge.get_adjustment_limits()
                row_units = matrix.get_units_from_row(wedge_used=self.wedge)
                if mat_units < row_units - limits.shrink:
                    # character too narrow for its row
                    space_units -= (row_units - mat_units - limits.shrink)
                    mat_units = row_units - limits.shrink
                elif mat_units > row_units + limits.stretch:
                    # character too wide for its row
                    space_units += (mat_units - row_units - limits.stretch)
                    mat_units = row_units + limits.stretch
                mat_wedges = self.get_wedge_positions(matrix, mat_units)
            # get these signals
            mat_codes = get_codes(matrix, mat_wedges)
            # omit spaces not wide enough
            if space_units >= 3:
                try:
                    # high spaces are preferred
                    space = self.find_space(space_units, low=False)
                except bm.MatrixNotFound:
                    # then we have to do with a low one...
                    space = self.find_space(space_units, low=True)
                sp_wedges = self.get_wedge_positions(space, space_units)
                space_codes = get_codes(space, sp_wedges)
            else:
                space_codes = []
                leftover_units += space_units
            # characters should be separated by at least one space
            return space_codes + mat_codes

        if self.diecase:
            self.display_diecase_layout()
        else:
            # select size for an empty layout
            self.resize_layout()

        if not UI.confirm('Proceed?', default=True, abort_answer=False):
            return

        queue = ['NJS 0005', 'NKJS 0075 0005']
        quad = self.quad.get_ribbon_record()
        pos_0075, pos_0005 = 3, 8
        leftover_units = 0
        # build the layout one by one
        for number, row in enumerate(self.diecase.layout.by_rows(), start=1):
            UI.display('Processing row {}'.format(number))
            queue.append(quad)
            for mat in row:
                queue.extend(add_matrix(mat))
            leftover_units = 0
            # single justification for line start
            queue.append(quad)
            queue.append('NKS 0075 {}'.format(pos_0075))
            queue.append('NKJS 0075 0005 {}'.format(pos_0005))

        return queue

    @mc.temp_wedge
    def calibrate_machine(self):
        """Casts the "en dash" characters for calibrating the character X-Y
        relative to type body."""
        def get_codes(char, default_position, comment):
            """Gets two mats for a given char and adjusts its parameters"""
            code = UI.enter('Where is the {}?'.format(char),
                            default=default_position)
            matrix = bm.Matrix(' ', code=code.upper())
            # try again recursively if wrong value
            if not matrix.position.row or not matrix.position.column:
                return get_codes(char, default_position, comment)
            # ask for unit width
            row_units = self.wedge[matrix.position.row]
            units = UI.enter('Unit width?', default=row_units)
            matrix.units = units
            # calculate the character width for measurement
            width = self.wedge.set_width / 12 * units / 18 * self.wedge.pica
            description = '{}, {:4f} inches wide'.format(comment, width)
            # calculate justifying wedge positions
            positions = mc.get_wedge_positions(matrix, self.wedge, units)
            pos_0075, pos_0005 = positions.pos_0075, positions.pos_0005
            use_s_needle = (pos_0075, pos_0005) != (3, 8)
            codes = matrix.get_ribbon_record(s_needle=use_s_needle,
                                             comment=description)
            sjust = ['NJS 0005 {}'.format(pos_0005),
                     'NKS 0075 {}'.format(pos_0075)]
            # use single justification to adjust character width, if needed
            return ([*sjust, codes, codes] if use_s_needle
                    else [codes, codes])

        UI.display('Mould blade opening and X-Y character calibration:\n'
                   'Cast G5, adjust the sort width to the value shown.\n'
                   '\nThen cast some lowercase "n" letters and n-dashes,\n'
                   'check the position of the character relative to the\n'
                   'type body and adjust the bridge X-Y.\n'
                   'Repeat if needed.\n')

        # build a casting queue and cast it repeatedly
        line_out = 'NKJS 0005 0075'
        pump_stop = 'NJS 0005'
        # use half-quad, quad, "n" and en-dash
        chars = [('half-quad', 'G5', 'full square'),
                 ('quad', 'O15', 'half square'),
                 ('n/h', None, 'serif overhanging test'),
                 ('dash or calibration mark', None, 'X-Y alignment test')]
        codes = (code for what in chars for code in get_codes(*what))
        sequence = [line_out, *codes, line_out, pump_stop]
        self.machine.cast(sequence)
        # G-8 calibration
        UI.display('\n\nCalibration done. Now adjust the matrix case draw rods'
                   '\nso that the diecase is not wobbling anymore.\n')
        with self.machine:
            self.machine.test_one('G8')
            UI.pause('Sending G8, press any key to stop...')

    def main_menu(self):
        """Main menu for the type casting utility."""
        def options():
            """Generate options based on current state of the program."""
            machine = self.machine
            is_casting = self.machine.is_casting()
            is_punching = not is_casting
            multiple_modes = len(self.machine.supported_operation_modes) >= 2

            got_ribbon = bool(self.ribbon)

            ret = [option(key='m', value=self.calibrate_machine, seq=5,
                          cond=is_casting,
                          text='Calibrate machine',
                          desc='Align the character width, then diecase'),

                   option(key='c', value=self.cast_composition, seq=10,
                          cond=is_casting and got_ribbon,
                          text='Cast composition',
                          desc='Cast type from a selected ribbon'),

                   option(key='r', value=self.choose_ribbon, seq=10,
                          text='Select ribbon',
                          desc='Select a ribbon from database or file'),

                   option(key='p', value=self.cast_composition, seq=30,
                          cond=is_punching and got_ribbon,
                          text='Punch ribbon',
                          desc='Punch a paper ribbon for casting'),

                   option(key='v', value=self.display_ribbon_contents, seq=80,
                          text='View codes', cond=got_ribbon,
                          desc='Display all codes in the selected ribbon'),

                   option(key='h', value=self.cast_material, seq=60,
                          cond=is_casting, text='Cast sorts or spaces',
                          desc='Cast characters from specified mats'),

                   option(key='q', value=self.cast_qr_code, seq=70,
                          cond=qrcode, text='Cast QR codes',
                          desc='Cast QR codes from high and low spaces'),

                   option(key='F3', value=self.choose_machine, seq=90,
                          text=('Change the machine in use (current: {})'
                                .format(machine))),

                   option(key='F4', value=self.machine.switch_operation_mode,
                          seq=91, cond=multiple_modes,
                          text=('Change operation mode (current: {})'
                                .format(machine.operation_mode))),

                   option(key='F5', value=self.display_details, seq=92,
                          text='Show details...',
                          desc='Display ribbon and interface information'),

                   option(key='F6', value=self.diecase_proof, seq=93,
                          text='Diecase proof',
                          desc='Cast every character from the diecase'),

                   option(key='F8', value=machine.diagnostics, seq=95,
                          text='Diagnostics menu...',
                          desc='Interface and machine diagnostic functions')]
            return ret

        header = ('rpi2caster - CAT (Computer-Aided Typecasting) '
                  'for Monotype Composition or Type and Rule casters.\n\n'
                  'This program reads a ribbon (from file or database) '
                  'and casts the type on a composition caster.'
                  '\n\nCasting / Punching Menu:')
        exceptions = (Finish, Abort, KeyboardInterrupt, EOFError)
        UI.dynamic_menu(options, header, catch_exceptions=exceptions)

    def display_details(self):
        """Collect ribbon, diecase and wedge data here"""
        data = [ribbon.parameters if self.ribbon else {},
                self.machine.parameters]
        UI.display_parameters(*data)
        UI.pause()



# -*- coding: utf-8 -*-
"""Static data for rpi2caster: unit arrangements, wedge definitions,
letter frequencies for different languages etc."""
from json import loads
from pkg_resources import resource_string as rs

SOURCE = 'rpi2caster.data'


def get_data(name):
    """Get the resource from a JSON-encoded file"""
    return loads(rs(SOURCE, '{}.json'.format(name)).decode())

# Unit arrangement dictionary: {UA_ID: {style1: {char1: unit_value1...}...}...}
UNIT_ARRANGEMENTS = get_data('unit_arrangements')
# Letter frequencies dictionary: {LANG: {char1: freq1...}...}
LETTER_FREQUENCIES = get_data('char_freqs')
# Wedge definitions: {WEDGE_SERIES: [row1_units, row2_units...]...}
WEDGE_DEFINITIONS = get_data('wedge_units')
WEDGE_ALIASES = ('10E: UK S527', '10L: UK S536', '11Q: UK S574',
                 '14E: UK S1406', '1A: UK S207', '1B: UK S209',
                 '1C: UK S210', '1O: UK S221', '1R: UK S223',
                 '1Q: UK S224', '2A: UK S233', '2Z: UK S261',
                 '3O: UK S275', '3Q: UK S277', '3Y: UK S286',
                 '4A: UK S295', '4G: UK S300', '5P: UK S327',
                 '5V: UK S371', '7J: UK S422', '7Z: UK S449',
                 '8A: UK S450', '8U: UK S409', 'TW: S535 typewriter',
                 'AK: EU S5', 'BO: EU S221', 'CZ: EU S336',
                 'A: UK S5', 'D: UK S46', 'E: UK S92',
                 'F: UK S94', 'G: UK S93', 'J: UK S101',
                 'K: UK S87', 'L: UK S96', 'M: UK S45',
                 'N: UK S88', 'O: UK S111', 'Q: UK S50',
                 'S: UK S197', 'V: UK S202', 'W: UK S205', 'X: UK S47')

# Typefaces by series - their names, styles
TYPEFACES = get_data('typefaces')
# Help and documentation
HELP = get_data('help')
# Easter egg
EASTER_EGG = """
    ╔═══════════════════════════════════════════════════════════════════════╗
    ║                                   _                              _    ║
    ║    /\___/\       __   __     __   |    __  _ __       __   __    |    ║
    ║   |       |     /  \ /  \   /  \ -+-  |  \ |/  \     /  \ /  \  -+-   ║
    ║  _  *   *  _   |      __ |  \__   |   +--/ |        |      __ |  |    ║
    ║  -   /_\   -   |     /  \|     \  |   |    |        |     /  \|  |    ║
    ║      ---        \__/_\__/\__\__/___\_/ \__/|_        \__/_\__/\___\_  ║
    ║                                                                       ║
    ║ Hello Kitty!              gives your MONOTYPE nine lives              ║
    ║                                                                       ║
    ╚═══════════════════════════════════════════════════════════════════════╝
    """



# -*- coding: utf-8 -*-
"""Typesetter program"""

from collections import OrderedDict
from contextlib import suppress
import itertools as it
import re

from .rpi2caster import CFG, DB, UI, Abort, option as opt
from . import basic_models as bm, basic_controllers as bc, definitions as d
from . import parsing as p
from .main_models import Ribbon
from .main_controllers import DiecaseMixin


# Ribbon control routines

@DB
def get_all_ribbons():
    """Lists all ribbons we have."""
    try:
        rows = Ribbon.select().order_by(Ribbon.ribbon_id)
        ribbons = OrderedDict(enumerate(rows, start=1))
        return ribbons
    except DB.OperationalError:
        Ribbon.create_table(fail_silently=True)
        return get_all_ribbons()


def list_ribbons(data=get_all_ribbons()):
    """Display all ribbons in a dictionary, plus an empty new one"""
    UI.display('\nAvailable ribbons:\n')
    template = ('|{no:<5}  {ribbon_id:<20} {diecase_id:<20} '
                '{wedge:<12} {customer:<20} {desc:<50}|')
    fields = dict(no='No.', ribbon_id='Ribbon ID', diecase_id='Diecase ID',
                  wedge='Wedge name', customer='Customer', desc='Description')
    UI.display_header(template.format(**fields))
    for index, ribbon in data.items():
        values = dict(no=index, desc=ribbon.description,
                      ribbon_id=ribbon.ribbon_id, diecase_id=ribbon.diecase_id,
                      wedge=ribbon.wedge.name, customer=ribbon.customer)
        UI.display(template.format(**values))


def ribbon_from_file(file=None, manual_choice=True):
    """Import the ribbon from file object specified as argument.
    Choose the file manually if file object is not specified.

    file: any object with the readlines() method"""
    ribbon_file = (file if 'readlines' in dir(file)
                   else open(file, 'r') if file
                   else UI.import_file() if manual_choice else None)
    if ribbon_file:
        ribbon = Ribbon()
        ribbon.import_from_file(ribbon_file)
        return ribbon


def choose_ribbon(fallback=Ribbon, fallback_description='new empty ribbon'):
    """Select ribbons from database and let the user choose one of them"""
    prompt = 'Your choice? (0 = {})'.format(fallback_description)
    data = get_all_ribbons()
    if not data:
        return fallback()
    UI.display('Choose a ribbon:', end='\n\n')
    list_ribbons(data)
    qty = len(data)
    # let the user choose the ribbon
    choice = UI.enter(prompt, default=0, datatype=int, minimum=0, maximum=qty)
    return data.get(choice) or fallback()


@DB
def get_ribbon(ribbon_id=None, fallback=choose_ribbon):
    """Get a ribbon with given ribbon_id"""
    if ribbon_id:
        try:
            return Ribbon.get(Ribbon.ribbon_id == ribbon_id)
        except Ribbon.DoesNotExist:
            UI.display('Ribbon {} not found in database.'.format(ribbon_id))
        except DB.OperationalError:
            Ribbon.create_table(fail_silently=True)
    return fallback()


class RibbonMixin:
    """Mixin for ribbon-related operations"""
    _ribbon = Ribbon()

    @property
    def ribbon(self):
        """Ribbon for the casting/typesetting session"""
        return self._ribbon

    @ribbon.setter
    def ribbon(self, ribbon):
        """Ribbon setter. False or None will result in new empty Ribbon()"""
        self._ribbon = ribbon or Ribbon()

    @ribbon.setter
    def ribbon_file(self, ribbon_file):
        """Use a ribbon file"""
        self.ribbon = ribbon_from_file(ribbon_file)

    @ribbon.setter
    def output_file(self, ribbon_file):
        """Use a ribbon file"""
        self.ribbon.file = ribbon_file

    @ribbon.setter
    def ribbon_id(self, ribbon_id):
        """Use a ribbon with a given ID, or an empty one"""
        self.ribbon = get_ribbon(ribbon_id, fallback=self.ribbon)

    def ribbon_by_name(self, name, choose_if_not_found=False):
        """Try to determine if the name supplied is a ribbon ID or filename.
        Otherwise use the fallback option."""
        # 1. try looking it up in the database
        # 2. try loading it from file
        # 3. optionally try choosing another ribbon if both lookups failed
        # 4. if all fails, just keep the current ribbon
        def sources():
            """try different ribbon sources"""
            # no name? fresh ribbon then
            if not name:
                yield Ribbon()
            # try looking it up in the database
            yield get_ribbon(name, fallback=lambda *_: None)
            # try to get a file
            yield ribbon_from_file(name, manual_choice=False)
            # choose manually
            if choose_if_not_found:
                yield choose_ribbon()
            # all of these failed = don't change anything
            yield self.ribbon

        for ribbon in sources():
            if ribbon:
                self.ribbon = ribbon
                break

    def choose_ribbon(self):
        """Chooses a ribbon from database or file"""
        with suppress(Abort):
            ribbon = choose_ribbon(fallback=ribbon_from_file,
                                   fallback_description='import from file')
            self.ribbon = ribbon

    def display_ribbon_contents(self, ribbon=None):
        """Displays the ribbon's contents, line after line"""
        UI.paged_display(ribbon or self.ribbon.contents, sep='\n')


class SourceMixin(object):
    """Mixin for source text"""
    _source = ''

    @property
    def source(self):
        """Source text for typesetting"""
        return self._source

    @source.setter
    def source(self, text):
        """Source setter"""
        self._source = text or ''

    @source.setter
    def text_file(self, text_file):
        """Use a file object as a source of text"""
        # If a string or None is passed as an argument,
        # AttributeError would be raised. We'd rather ignore it.
        with suppress(AttributeError), text_file:
            self.source = ''.join(text_file.readlines())

    def edit_text(self):
        """Edits the input text"""
        self.source = UI.edit(self.source)
        return self.source


class TypesettingContext(SourceMixin, DiecaseMixin, RibbonMixin):
    """Mixin for setting diecase, wedge and measure"""
    _measure, _default_alignment = None, d.ALIGNMENTS.left
    # markup format used for source text
    markup_language = 'enkiduml'
    manual_mode = False

    @property
    def measure(self):
        """Typesetting measure i.e. line length"""
        if not self._measure:
            # get measure configuration and return default one
            measure = CFG['Typesetting'].get('default_measure')
            unit = CFG['Typesetting'].get('measurement_unit')
            self._measure = bm.Measure(measure, unit, self.wedge.set_width)
        return self._measure

    @measure.setter
    def measure(self, measure):
        """Measure setter"""
        with suppress(ValueError):
            unit = CFG['Typesetting'].get('measurement_unit')
            self._measure = bm.Measure(measure, unit, self.wedge.set_width)

    @measure.setter
    def line_length(self, measure):
        """Set the line length for typesetting"""
        self.measure = measure

    @property
    def default_alignment(self):
        """Default alignment:
        Determines how paragraphs ending with a double newline ("\n\n")
        and the end of the source text will be aligned.
        Valid options: "left", "right", "center", "both".
        """
        return self._default_alignment

    @default_alignment.setter
    def default_alignment(self, alignment):
        """Default alignment setter"""
        options = {}
        for alignment in d.ALIGNMENTS:
            shorts = [alignment.name, alignment.name[0],
                      *(code.strip('^') for code in alignment.codes)]
            options.update({s: alignment for s in shorts})

        string = alignment.strip().replace('^', '').lower()
        value = options.get(string)
        if value:
            self._default_alignment = value

    def change_measure(self):
        """Change a line length"""
        old_meas = self.measure
        measure = bc.set_measure(old_meas, what='line length / galley width')
        self.measure = measure

    def change_alignment(self):
        """Changes the default text alignment"""
        msg = 'Choose default alignment for paragraphs:'
        opts = [opt(key='l', value=d.ALIGNMENTS.left, text='left', seq=1),
                opt(key='c', value=d.ALIGNMENTS.center, text='center', seq=2),
                opt(key='r', value=d.ALIGNMENTS.right, text='right', seq=3),
                opt(key='b', value=d.ALIGNMENTS.both, text='justify', seq=4)]
        align = UI.simple_menu(msg, opts, default_key='b', allow_abort=True)
        self.default_alignment = align

    def toggle_manual_mode(self):
        """Changes the manual/automatic typesetting mode"""
        self.manual_mode = not self.manual_mode


class Document:
    """A document represents a whole page of text to be cast in one run.
    It may and probably will consist of multiple paragraphs,
    separated by double newlines ('\n\n')."""
    def __init__(self, text, context):
        self.text = str(text)
        self.context = context
        # split the text into paragraphs with text and alignment
        parsed = self.make_paragraphs(self.text)
        self.paragraphs = [Paragraph(t, self.context, a) for t, a in parsed]

    def compose(self):
        """Typeset the document. The context is a TypesettingContext
        and stores information such as line length, alignments etc."""
        for paragraph in self.paragraphs:
            paragraph.compose()

    def translate(self):
        """Translate the document into Monotype codes"""
        return sum((par.translate() for par in self.paragraphs), [])

    def make_ribbon(self):
        """Make a ribbon i.e. signals sequence out of the document"""
        sequence = (obj.get_ribbon_record() for obj in self.translate())
        contents = '\n'.join(sequence)
        ribbon = Ribbon()
        ribbon.contents = contents
        self.contxt.ribbon = ribbon

    def make_paragraphs(self, text):
        """Cut a text into paragraphs and add them to self.paragraphs"""
        def_alignment = self.context.default_alignment
        align_codes = {code: jc for jc in d.ALIGNMENTS for code in jc.codes}
        # regex to catch and handle it
        codes_regex = r'(?<=\^)({})'.format('|'.join(align_codes.keys()))
        # double newline acts as a paragraph break with default justification
        align_codes['\n\n'] = def_alignment
        # triple newline means "stop parsing, discard everything further"
        align_codes['\n\n\n'] = def_alignment
        # partial regular expression to catch the newlines
        newlines_regex = r'\n{2, 3}'
        # regex to catch both codes and newlines
        regex = re.compile('({})|({})'.format(newlines_regex, codes_regex),
                           flags=re.IGNORECASE)
        # split the string with regex, discard empty strings
        parts = [align_codes.get(s, s) for s in regex.split(text) if s]
        ret = list(it.zip_longest(*[iter(parts)] * 2, fillvalue=def_alignment))
        print(ret)
        return ret

    def _cut_to_paragraphs(self, text):
        """Cut the incoming input string into paragraphs.
        A paragraph delimiter can be:
            -a justification control code,
            -a double newline.
        Only the part of text before "end of file" control code,
        or triple newline, is processed."""
        default_alignment = self.context.default_alignment
        da_token = ('DefaultAlignment', '\n{2}', default_alignment)
        # code capturing group
        capture_group = r'(?<=\^){}'

        # token definitions: end-of-file, double newline -> default alignment
        tds = [('EOF', r'(\n{3}|(?<=\^)ef)', default_alignment), da_token]
        # justification code token definitions
        tds.extend([(a.name, capture_group.format('|'.join(a.codes)), a)
                    for a in d.ALIGNMENTS])
        tds.append(('Break', re.escape('^cr')))
        tds.append(('Text', r'.+'))

        # get a stream of tokens in the text
        valid_tokens = []
        tokens = p.tokenize(text, tds)
        for token in tokens:
            valid_tokens.append(token)
            if token.name == 'EOF':
                break
        zipped = it.zip_longest(*[iter(valid_tokens)] * 2, fillvalue=da_token)
        ret = [(t.value, c.routine) for (t, c) in zipped]
        print(ret)
        return ret


class Paragraph(object):
    """A page is broken into a series of paragraphs."""
    def __init__(self, text, context, alignment=None):
        self.text = text
        self.context = context
        self.alignment = alignment or context.default_alignment
        # markup parsing routine
        parsers = {'enkiduml': self.parse_enkiduml}
        parser = parsers.get(self.context.markup_language)
        self.tokens = parser(text)
        self.lines = []

    def display_text(self):
        """Prints the text"""
        UI.display(self.text)

    def parse_enkiduml(self, text):
        """Typeset the paragraph's text.
        Input: text with EnkiduML codes,
        output: tuples [(char, style, unit_correction), ...]"""
        def add_units(pattern):
            """Adds a specified number of units to a character"""
            nonlocal unit_correction
            unit_correction += p.get_number(pattern)

        def subtract_units(pattern):
            """Takes away a specified number of quarters of a unit
            from a character"""
            nonlocal unit_correction
            unit_correction -= p.get_number(pattern) / 4.0

        def add_quads(pattern):
            """Add some quads to the composition"""
            self.current_line.extend(['1em'] * p.get_number(pattern))

        def add_enquads(pattern):
            """Add some half-quads to the composition"""
            self.current_line.extend(['0.5em'] * p.get_number(pattern))

        def change_style(pattern):
            """Change the current style"""
            self.context.current_style = d.STYLE_COMMANDS.get(pattern)

        codes = [p.make_number_pattern('#=', 1, 3),
                 p.make_number_pattern('+-', 1, 2),
                 *d.STYLE_COMMANDS.keys()]

        unit_correction = 0
        # tokens = p.split_on_matches(text, codes=codes)
        # print(tokens)


    def translate(self):
        """Translate the characters into a series of Monotype codes"""



# coding: utf-8
"""Diecase manipulation functions and classes"""

import csv
from collections import OrderedDict
from contextlib import suppress
from copy import copy
from functools import wraps

from . import basic_models as bm, basic_controllers as bc, definitions as d
from . import views
from .rpi2caster import USER_DATA_DIR, UI, Abort, Finish, option
from .data import WEDGE_DEFINITIONS
from .main_models import DB, Diecase, Wedge, TypefaceSize, UnitArrangement


def import_csv(diecase, filename=''):
    """Imports a layout from file"""
    def preview(num=5):
        """Preview the first five records in the file"""
        with layout_file:
            records = csv.reader(layout_file, line_num=num,
                                 delimiter=';', quotechar='"')
            text = '\n'.join(''.join(str(field).ljust(15) for field in rec)
                             for rec in records)
            UI.display('File preview: displaying first {} rows:\n'
                       .format(num))
            UI.display(text)

    # Load the layout from file
    try:
        layout_file = UI.import_file(filename)
        preview()
        with layout_file:
            csv_reader = csv.reader(layout_file, delimiter=';', quotechar='"')
            layout_gen = (record for record in csv_reader)
        if UI.confirm('Is the 1st row a table header? ', default=True):
            next(layout_gen)
        if not UI.confirm('Proceed?', default=True, abort_answer=False):
            return
        # parse the layout
        diecase.load_layout(rec for rec in layout_gen)
    except (TypeError, ValueError):
        UI.pause('Incorrect layout. Check your file.', allow_abort=True)


def export_csv(diecase, filename=''):
    """Exports the matrix case layout to file."""
    name = diecase.diecase_id or 'NewDiecase'
    # Save the exported diecase layout in the default directory
    file_name = filename or '{}/{}.csv'.format(USER_DATA_DIR, name)
    with suppress(Abort):
        output_file = UI.export_file(file_name)
        with output_file:
            csv_writer = csv.writer(output_file, delimiter=';',
                                    quotechar='"', quoting=csv.QUOTE_ALL)
            csv_writer.writerow(['Char', 'Styles', 'Position', 'Units'])
            # store mats with no position info as well
            csv_writer.writerows(diecase.get_raw_layout())
        UI.pause('File {} successfully saved.'.format(filename))


def edit_diecase_typeface_data(diecase):
    """Edit the typeface data for a diecase"""
    def set_typefaces():
        """Assign styles to typeface. Will also look for related typefaces."""
        ua_prompt = 'Unit arrangement variant? Possible options:\n{}'
        for style in styles:
            try:
                # what was set up previously?
                # (helpful when choosing the typeface/UA manually)
                old_typeface, old_unit_arrangement = typeface_data[style]
                old_typeface_series = old_typeface.series
                old_ua_number = old_unit_arrangement.number
            except (KeyError, ValueError, AttributeError):
                old_ua_number, old_typeface_series = '', ''

            try:
                # auto-assign typeface
                variant = main_typeface.get_variant(style)
                if not variant:
                    raise ValueError
            except ValueError:
                # assign manually
                _num = UI.enter('Typeface number for {}?'.format(style.name),
                                default=old_typeface_series)
                _typeface = TypefaceSize(_num, size)
                # choose only one style here
                _style = bc.choose_styles(_typeface.styles.first,
                                          multiple=False,
                                          mask=_typeface.styles).first
                variant = _typeface.get_variant(_style)

            # try to automatically select unit arrangement from typeface data
            # if this fails, choose the unit arrangement manually
            unit_arrangement = variant.unit_arrangement
            if not unit_arrangement:
                views.list_unit_arrangements()
                UI.display('Cannot automatically choose the unit arrangement.')
                ua_number = UI.enter('Unit arrangement number?',
                                     default=old_ua_number)
                uarr = UnitArrangement(ua_number)
                ua_variant = UI.enter(ua_prompt.format(uarr.variant_names),
                                      default='r')
                unit_arrangement = uarr.get_variant(ua_variant)
            typeface_data[style] = (variant, unit_arrangement)

    def assign_normal_wedge():
        """Assign a normal wedge for the diecase"""
        wedge = diecase.wedge
        series = wedge.series
        set_width = main_typeface.set_width
        suffix = 'E' if wedge.is_brit_pica else ''
        designation = 'S{}-{}{}'.format(series, set_width, suffix)
        diecase.wedge = choose_wedge(designation)

    def set_diecase_id():
        """Set the diecase ID (unique name) for this diecase."""
        faces = [item[0] for item in typeface_data.values()]
        size = main_typeface.size
        numbers = '+'.join(sorted({tf.series for tf in faces}))
        UI.display('Enter a name used for identifying your diecase.\n'
                   'It MUST be unique and can, but does not have to, '
                   'include the typeface series and size.\n'
                   'Customize the name to your preference.')
        suggested = '{}-{}-'.format(numbers, size)
        diecase.diecase_id = UI.enter('Diecase ID?', default=suggested)

    typeface_data = diecase.typeface_data
    # get the styles currently assigned to the diecase
    styles = bm.Styles(typeface_data)
    # get the main typeface, most often but not always roman
    # if this fails, get the variant for the Times New Roman 12D roman
    try:
        # changing the data already configured
        main_typeface = typeface_data.get(styles.first)[0]
        if not main_typeface:
            raise ValueError
    except (IndexError, ValueError):
        # no data entered yet
        main_typeface = TypefaceSize()
    UI.display('Diecase: {}'.format(diecase.description))
    UI.display('Editing the typeface information:\n'
               'First choose the typeface series and size.\n'
               'New diecases have 327-12D roman prefilled by default, '
               'change the number and size to your need.\n')
    while True:
        # edit series and size
        series = UI.enter('Type series?', main_typeface.series).upper()
        size = UI.enter('Type size?', main_typeface.size).upper()
        main_typeface = TypefaceSize(series, size)
        # edit or update styles
        styles = bc.choose_styles(styles, mask=main_typeface.combined_styles)
        # choose typeface per style, automatically or manually
        # choose unit arrangements for each style, and normal wedge for diecase
        set_typefaces()
        assign_normal_wedge()
        if UI.confirm('All OK? If not, start again.', allow_abort=True):
            break

    diecase.typeface_data = typeface_data
    if not diecase.diecase_id:
        set_diecase_id()


@DB
def get_all_diecases():
    """Lists all matrix cases we have."""
    try:
        rows = Diecase.select().order_by(Diecase.diecase_id)
        enumerated_diecases = enumerate(rows, start=1)
        return OrderedDict(enumerated_diecases)
    except Diecase.DoesNotExist:
        return {}
    except DB.OperationalError:
        Diecase.create_table(fail_silently=True)
        return get_all_diecases()


@DB
def count_diecases():
    """Count the diecases in database"""
    try:
        return Diecase.select().count()
    except DB.OperationalError:
        # since we're creating a new table, we're sure no diecases are there
        Diecase.create_table(fail_silently=True)
        return 0


@DB
def check_persistence(diecase_id):
    """Check if the diecase is stored in the database"""
    try:
        query = Diecase.select().where(Diecase.diecase_id == diecase_id)
        return query.count()
    except DB.OperationalError:
        Diecase.create_table(fail_silently=True)
        return 0


def choose_diecase(fallback=Diecase, fallback_description='new empty diecase'):
    """Select diecases from database and let the user choose one of them.
    If no diecases are found, return None and let the calling logic
    determine what fallback to use."""
    prompt = 'Your choice? (0 = {})'.format(fallback_description)
    data = get_all_diecases()
    if not data:
        return fallback().load()
    UI.display('Choose a matrix case:', end='\n\n')
    views.list_diecases(data)
    qty = len(data)
    # let the user choose the diecase
    choice = UI.enter(prompt, default=0, datatype=int, minimum=0, maximum=qty)
    diecase = data.get(choice) or fallback()
    return diecase.load()


@DB
def get_diecase(diecase_id=None, fallback=choose_diecase):
    """Get a diecase with given parameters"""
    if diecase_id:
        try:
            return Diecase.get(Diecase.diecase_id == diecase_id).load()
        except Diecase.DoesNotExist:
            UI.display('Diecase {} not found in database!'.format(diecase_id))
        except DB.OperationalError:
            Diecase.create_table(fail_silently=True)
    return fallback().load()


def temp_diecase(routine):
    """Use a temporary diecase"""
    @wraps(routine)
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        if not self.diecase:
            # empty diecase? then let the user choose...
            old_diecase, self.diecase = self.diecase, choose_diecase()
            UI.display_parameters(self.diecase.parameters)
            UI.display('\n\n')
            retval = routine(self, *args, **kwargs)
            # cleanup: restore the previous diecase
            self.diecase = old_diecase
        else:
            # if diecase was chosen before, don't ask
            retval = routine(self, *args, **kwargs)
        return retval
    return wrapper


# Diecase layout controller routines


def resize_layout(diecase):
    """Change the diecase layout size"""
    # select one of 3 sizes used by Monotype
    sizes = [(15, 15), (15, 17), (16, 17)]
    options = [option(key=n, value=size, text='{} x {}'.format(*size))
               for n, size in enumerate(sizes, start=1)]
    selected_size = UI.simple_menu(message='Matrix case size:',
                                   options=options,
                                   default_key=2, allow_abort=True)
    diecase.resize(*selected_size)


def edit_layout(diecase):
    """Edits a matrix case layout, row by row, matrix by matrix.
    Allows to enter a position to be edited. """
    def swap():
        """Swap two matrices based on command"""
        mat1_code = UI.enter('Matrix position to replace?')
        mat1 = diecase.select_one(code=mat1_code)
        mat2_code = UI.enter('Other matrix position?')
        mat2 = diecase.select_one(code=mat2_code)
        # no exceptions? then swap
        mat1.code, mat2.code = mat2.code, mat1.code

    def edit(mat, single=False):
        """Edit a matrix"""
        views.display_layout(diecase)
        mat = bc.edit_matrix(mat, single=single)

    def all_rows():
        """Row-by-row editing - all cells in row 1, then 2 etc."""
        sequence = [mat for column_number in diecase.column_numbers
                    for mat in diecase.select_column(column_number)]
        for mat in sequence:
            edit(mat)

    def all_columns():
        """Column-by-column editing - all cells in column NI, NL, A...O"""
        sequence = [mat for row_number in diecase.row_numbers
                    for mat in diecase.select_row(row_number)]
        for mat in sequence:
            edit(mat)

    def single_row():
        """Edits matrices found in a single row"""
        row = UI.enter('Row?', datatype=int, default='',
                       minimum=1, maximum=diecase.rows)
        for mat in diecase.select_row(row):
            edit(mat)

    def single_column():
        """Edits matrices found in a single column"""
        def condition(col_num):
            """Validation condition for column number"""
            return col_num.upper() in diecase.column_numbers

        prompt = 'Column? [{}]'.format(', '.join(diecase.column_numbers))
        column = UI.enter(prompt, default='', condition=condition).upper()
        for mat in diecase.select_column(column):
            edit(mat)

    def single_matrix():
        """Edits a single matrix with specified coordinates"""
        position = UI.enter('Coordinates?', default='')
        mat = diecase.select_one(code=position.upper())
        edit(mat, single=True)

    def show_layout():
        """Shows diecase layout and pauses"""
        UI.display('\nCurrent diecase layout:\n')
        views.display_layout(diecase)
        UI.pause()

    def options():
        """Menu options"""
        ret = [option(key='r', value=single_row, text='Single row', seq=10),
               option(key='R', value=all_rows,
                      text='All mats, row by row', seq=20),
               option(key='c', value=single_column,
                      text='Single column', seq=30),
               option(key='C', value=all_columns,
                      text='All mats, column by column', seq=40),
               option(key='m', value=single_matrix,
                      text='Single matrix', seq=5),
               option(key='s', value=swap, text='Swap two matrices', seq=60),
               option(key='l', value=show_layout,
                      text='View current layout', seq=70),
               option(key='Ins', value=diecase.store_layout,
                      text='Save the changes in diecase layout', seq=80),
               option(key='Esc', value=Finish, text='Finish editing', seq=99)]
        return ret

    while True:
        opt = UI.simple_menu('Diecase layout edition menu:', options=options)
        with suppress(TypeError):
            raise opt
        with suppress(Abort, Finish, bm.MatrixNotFound):
            opt()


def test_layout_charset(diecase):
    """Tests completeness for a chosen language or text."""
    def get_lang_chars(lang):
        """Get an ordered set of characters in a language"""
        characters = bm.CharFreqs(lang)
        lowercase = (char for char in characters)
        uppercase = (char.upper() for char in characters)
        return lambda: (*uppercase, *lowercase)

    def enter_text():
        """Get an ordered set of all characters found in a text"""
        text = UI.edit()
        return sorted(set(char for char in text if not char.isspace()))

    def make_charset():
        """Use a menu to choose language or enter custom text,
        then return a set of all characters found"""
        lang_options = [option(value=get_lang_chars(lang),
                               text='{} - {}'.format(lang, lang_name))
                        for lang, lang_name in sorted(d.LANGS.items())]
        options = [*lang_options,
                   option(key='Esc', value=Abort, seq=99, text='Exit'),
                   option(key='e', value=enter_text, seq=98,
                          text='Custom text')]
        charset = UI.simple_menu('Choose language or enter text',
                                 options, default_key='e', allow_abort=False)()
        return sorted(set(charset))

    def find_missing_mats(style):
        """Look up characters of a given style in the diecase layout"""
        missing = []
        for char in charset:
            mat = lookup_table.get((char, style))
            if not mat:
                missing.append(char)
        return missing

    lookup_table = diecase.get_lookup_table()
    styles = bc.choose_styles(diecase.styles)
    charset = make_charset()
    # which characters we don't have, grouped by style
    checks = {style: find_missing_mats(style) for style in styles}
    missing_by_style = {style: missing
                        for style, missing in checks.items() if missing}

    # if we have all needed characters, all missing char lists are empty
    # the diecase is complete, so return True
    if not missing_by_style:
        UI.pause('All characters are present.')
        return True

    # info for user
    info = 'Missing matrices for {}: {}'

    # sort the missing mats by style, then by char
    for style in styles:
        missing_chars = missing_by_style.get(style)
        if not missing_chars:
            continue
        missing_chars_string = ', '.join(sorted(missing_chars))
        UI.display(info.format(style.name, missing_chars_string))

    # the diecase is incomplete
    UI.pause()
    return False


def find_matrix(diecase, choose=True, **kwargs):
    """Search the diecase layout and get a matching mat.

    char, styles, position, units: search criteria,
    choose: manual choice menu or new mat definition."""
    def define_new_mat():
        """Create a new Matrix object"""
        position = (kwargs.get('position', '') or kwargs.get('pos', '') or
                    UI.enter('Matrix coordinates?', default=''))
        if not position:
            raise bm.MatrixNotFound('No coordinates entered')
        mat = bm.Matrix(char=char, styles=styles, code=position,
                        diecase=diecase)
        return mat

    def choose_from_menu():
        """Display a menu to choose mats"""
        matrices = sorted(mats, key=lambda mat: (mat.char, mat.position))
        menu_data = {i: mat for i, mat in enumerate(matrices, start=1)}
        # no matches? make a new one!
        if not menu_data:
            return None

        # build title depending on char and styles
        st_string = ('' if styles.use_all or not char
                     else '{}'.format(styles.names))
        title = ['Multiple matrices', 'for' if char else '',
                 st_string, char, 'found. Please choose a matrix:\n']
        UI.display(' '.join(s for s in title if s))

        # table header
        row = '{:<7}{}'
        UI.display_header(row.format('Index', 'Matrix'))

        # show available matrices
        for i, mat in menu_data.items():
            UI.display(row.format(i, mat))
        UI.display()

        # let user choose
        choice = UI.enter('Choose matrix (0 or blank to enter manually)',
                          default=0, minimum=0, maximum=len(menu_data))
        # if mat is None, define a new one
        return menu_data.get(choice)

    char = kwargs.get('char', '')
    styles = bm.Styles(kwargs.get('styles', '*'))
    mats = diecase.select_many(**kwargs)
    if len(mats) == 1:
        # only one match: return it
        return mats[0]
    elif choose:
        # manual mat choice
        return choose_from_menu() or define_new_mat()
    elif mats:
        # automatic choice, multiple matches found: get first available
        return mats[0]
    else:
        raise bm.MatrixNotFound('Automatic matrix lookup failed')


# Wedge controller routines

def choose_wedge(wedge_name=None):
    """Choose a wedge manually"""
    def enter_name():
        """Enter the wedge's name"""
        # List known wedges
        views.list_wedges()
        return UI.enter('Wedge designation?', default=default_wedge)

    def enter_parameters(name):
        """Parse the wedge's name and return a list:
        [series, set_width, is_brit_pica, units]"""
        def divisible_by_quarter(value):
            """Check if a value is divisible by 0.25:
            1, 3.0, 1.25, 2.5, 5.75 etc. -> True
            2.2, 6.4 etc. -> False"""
            return not value % 0.25

        # For countries that use comma as decimal delimiter, convert to point:
        w_name = name.replace(',', '.').upper().strip()
        # Check if this is an European wedge
        # (these were based on pica = .1667" )
        is_brit_pica = w_name.endswith('E')
        # Away with the initial S, final E and any spaces before and after
        # Make it work with space or dash as delimiter
        # ("S5-12" and "S5 12" should work the same)
        wedge = w_name.strip('SE ').replace('-', ' ').split(' ')
        try:
            series, set_width = wedge
        except ValueError:
            series, set_width = wedge, 0
        # Now get the set width - ensure that it is float divisible by 0.25
        # no smaller than 5 (narrowest type), no wider than 20 (large comp)
        prompt = ('Enter the set width as a decimal fraction '
                  'divisible by 0.25 - e.g. 12.25: ')
        set_width = UI.enter(prompt, datatype=float)

        set_width = UI.enter(prompt, default=set_width, datatype=float,
                             minimum=5, maximum=20,
                             condition=divisible_by_quarter)
        # We have the wedge name, so we can look the wedge up in known wedges
        # (no need to enter the unit values manually)
        current_units = WEDGE_DEFINITIONS.get(series, d.S5)
        prompt = ('Enter the wedge unit values for rows 1...15 '
                  'or 1...16, separated by commas.\n')
        units = UI.enter(prompt, default=current_units, minimum=15, maximum=16)
        # Now we have the data...
        return {'series': series, 'set_width': set_width,
                'is_brit_pica': is_brit_pica, 'units': units}

    default_wedge = str(wedge_name) if wedge_name else 'S5-12E'
    w_name = enter_name()
    try:
        return Wedge(wedge_name=w_name)
    except ValueError:
        data = enter_parameters(w_name)
        return Wedge(wedge_data=data)


def temp_wedge(routine):
    """Decorator for typesetting and casting routines.
    Assign a temporary alternative wedge for casting"""
    @wraps(routine)
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        # Assign a temporary wedge
        old_wedge, self.wedge = self.wedge, choose_wedge(self.wedge.name)
        UI.display_parameters(self.wedge.parameters)
        UI.display('\n\n')
        retval = routine(self, *args, **kwargs)
        # Restore the former wedge and exit
        self.wedge = old_wedge
        return retval
    return wrapper


def get_wedge_positions(matrix, normal_wedge, units, correction=0):
    """Calculate the 0075 and 0005 wedge positions for this matrix
    based on the current wedge used.

    matrix - a Matrix object
    correction - units of self.wedge's set to add/subtract,
    units - arbitrary character width in units of self.wedge's set"""
    def steps(wedge, unit_width=0):
        """get a width (in .0005" steps) of a character
        for a given number of units or diecase row"""
        inches = unit_width / 18 * wedge.set_width / 12 * wedge.pica
        # return a number of 0.0005" steps
        return int(2000 * inches)

    def limits_exceeded():
        """raise an error if width can't be adjusted with wedges"""
        limits = normal_wedge.get_adjustment_limits(matrix.islowspace())
        minimum = row_units - limits.shrink
        maximum = row_units + limits.stretch
        message = ('{}: desired width of {} units exceeds '
                   'adjustment limits (min: {} / max: {})')
        width = units + delta
        error_msg = message.format(matrix, width, minimum, maximum)
        raise bm.TypesettingError(error_msg)

    # absolute width: how many .0005" steps is it?
    char_width = steps(normal_wedge, units)
    # how many do we need to add or take away? (for kerning etc.)
    delta = steps(normal_wedge, correction)
    # how wide would a character from the given row normally be?
    # (using self.wedge, not self.diecase.wedge!)
    row_units = normal_wedge[matrix.position.row]
    row_width = steps(normal_wedge, row_units)
    # calculate the difference and wedge positions
    # 1 step of 0075 wedge is 15 steps of 0005; neutral positions are 3/8
    # 3 * 15 + 8 = 53, so any increment/decrement is relative to this
    increments = char_width + delta - row_width + 53
    # Upper limit: 15/15 => 15*15=225 + 15 = 240;
    # lower limit:  1/ 1 => 1 * 15 + 1 = 16
    if increments < 16 or increments > 240:
        limits_exceeded()
    # calculate wedge positions from the increments
    pos_0005, pos_0075 = increments % 15, increments // 15
    if not pos_0005:
        # wedge positions cannot be zero
        pos_0005, pos_0075 = 15, pos_0075 - 1
    return d.WedgePositions(pos_0075, pos_0005)


class DiecaseMixin:
    """Mixin for diecase-related operations"""
    _wedge, _diecase = None, None

    @property
    def wedge(self):
        """Get the temporary wedge, or the diecase's assigned wedge"""
        selected, fallback = self._wedge, self.diecase.wedge
        return selected if selected else fallback if fallback else Wedge()

    @wedge.setter
    def wedge(self, wedge):
        """Set the temporary wedge"""
        self._wedge = wedge

    @wedge.setter
    def wedge_name(self, wedge_name):
        """Set the wedge with a given name"""
        if not wedge_name:
            return
        try:
            self.wedge = Wedge(wedge_name=wedge_name)
        except ValueError:
            # parsing failed
            self.wedge = choose_wedge(wedge_name)

    def choose_wedge(self):
        """Chooses a new wedge"""
        self.wedge = choose_wedge(self.wedge)
        return self.wedge

    def get_units(self, matrix):
        """Get the number of units translated to currently used wedge.
        Use this for calculating unit width in typesetting routines."""
        dc_wedge = self.diecase.wedge
        width = matrix.units * dc_wedge.set_width * dc_wedge.pica
        # self.wedge units
        return width / self.wedge.set_width / self.wedge.pica

    def get_wedge_positions(self, matrix, units=0, correction=0):
        """Calculate the 0075 and 0005 wedge positions for this matrix
        based on the current wedge used.

        matrix - a Matrix object
        correction - units of self.wedge's set to add/subtract,
        units - arbitrary character width in units of self.wedge's set"""
        char_units = units or self.get_units(matrix)
        return get_wedge_positions(matrix, self.wedge, char_units, correction)

    @property
    def diecase(self):
        """Get a diecase or empty diecase, lazily instantiating a new one
        if none was chosen before"""
        diecase = self._diecase
        if diecase is None:
            # instantiate a new one and cache it
            diecase = Diecase().load()
            self._diecase = diecase
        return diecase

    @diecase.setter
    def diecase(self, diecase):
        """Set a diecase; keep the wedge"""
        self._diecase = diecase

    @property
    def diecase_id(self):
        """Get the diecase ID"""
        return self.diecase.diecase_id

    @diecase_id.setter
    def diecase_id(self, diecase_id):
        """Set a diecase with a given diecase ID, or a current/default one"""
        self.diecase = get_diecase(diecase_id, fallback=self.diecase)

    def choose_diecase(self):
        """Chooses a diecase from database"""
        self.diecase = choose_diecase(fallback=self.diecase,
                                      fallback_description='keep current')
        return self.diecase

    @property
    def charset(self):
        """Get a {style: {char: Matrix object}} charset from the diecase"""
        return self.diecase.get_charset()

    @property
    def space(self):
        """Get a space from diecase; most typically G2, 6-units wide"""
        return self.find_space(units=6, low=True)

    @property
    def half_quad(self):
        """Get a 9-unit quad (half-square) from diecase"""
        return self.find_space(units=9, low=True)

    @property
    def quad(self):
        """Get a full em quad"""
        return self.find_space(units=18, low=True)

    def find_space(self, units, low=True):
        """Find a matching space. If unit width is specified, try to
        look automatically, otherwise let user decide how wide it should be."""
        if not units:
            width = bc.set_measure(input_value='9u', unit='u',
                                   what='space width',
                                   set_width=self.wedge.set_width)
            units = width.units

        return self.diecase.get_space(units, low, wedge=self.wedge)

    def find_matrix(self, choose=True, **kwargs):
        """Search the diecase layout and get a matching mat.

        choose: manual choice menu or new mat definition

        kwargs: char, styles, position, units: search criteria."""
        return find_matrix(self.diecase, choose, **kwargs)

    def resize_layout(self):
        """Resize the layout of currently used diecase"""
        resize_layout(self.diecase)
        self.diecase.store_layout()

    @UI.paused
    def display_diecase_layout(self):
        """Display the diecase layout, unit values, styles."""
        views.display_layout(self.diecase)

    def test_diecase_charset(self):
        """Test whether the diecase layout has all required characters,
        either for a text or language's character set."""
        test_layout_charset(self.diecase)

    def diecase_manipulation(self):
        """A menu with all operations on a diecase"""
        @DB
        def _save():
            """Stores the matrix case definition/layout in database"""
            is_stored = check_persistence(self.diecase.diecase_id)
            self.diecase.store()
            self.diecase.save(force_insert=not is_stored)
            UI.pause('Data saved.')

        @DB
        def _save_as():
            """Stores the diecase in database with a different diecase ID"""
            diecase = copy(self.diecase)
            while check_persistence(diecase.diecase_id):
                prompt = 'New diecase ID (must be unique)?'
                diecase.diecase_id = UI.enter(prompt, default=diecase)
            diecase.save(force_insert=True)

        @DB
        def _delete():
            """Deletes a diecase from database"""
            prompt = 'Are you sure?'
            ans = UI.confirm(prompt, default=False, abort_answer=False)
            if ans:
                self.diecase.delete_instance()
                UI.pause('Matrix case deleted.')

        def _change_diecase():
            """Chooses another diecase"""
            self.diecase = choose_diecase()

        def _edit_typeface():
            """Edit diecase's typeface info"""
            edit_diecase_typeface_data(self.diecase)
            if UI.confirm('Save the diecase in database?'):
                _save()

        def _display_arrangements():
            """Display all unit arrangements for this diecase"""
            for unit_arrangement in self.diecase.unit_arrangements.values():
                views.display_ua_variant(unit_arrangement)
                UI.pause()

        def _edit_layout():
            """Edits a matrix case layout, row by row, matrix by matrix.
            Allows to enter a position to be edited. """
            edit_layout(self.diecase)

        def _clear_layout():
            """Generates a new layout for the diecase"""
            if UI.confirm('Are you sure?', default=False, abort_answer=False):
                self.diecase.purge()

        def _import():
            """Import diecase layout from CSV"""
            return import_csv(self.diecase)

        def _export():
            """Export diecase layout to CSV"""
            return export_csv(self.diecase)

        def header():
            """Menu header"""
            if not self.diecase.diecase_id:
                return ('Diecase manipulation menu:\n'
                        'To start, you must name your diecase.\n'
                        'The name must be unique, for example "327-12-01".')
            elif not check_persistence(self.diecase.diecase_id):
                return ('Diecase manipulation menu:\n'
                        'Diecase ID: {}\n'
                        'The diecase is not yet stored in the database.\n'
                        'You have to save it first before you can edit it.'
                        .format(self.diecase.diecase_id))
            return ('Diecase manipulation menu:\n\nWorking on {} ({})'
                    .format(self.diecase.diecase_id, self.diecase.description))

        def options():
            """Generate menu options"""
            is_stored = check_persistence(self.diecase.diecase_id)
            diecases_present = count_diecases()
            ret = [option(key='l', value=self.display_diecase_layout, seq=1,
                          text='Display diecase layout'),
                   option(key='e', value=_edit_layout,
                          text='Edit diecase layout', seq=2),
                   option(key='u', value=_display_arrangements, seq=4,
                          text='Display unit arrangements for this diecase'),
                   option(key='i', value=_edit_typeface, seq=11,
                          text='Edit the diecase information'),
                   option(key='t', value=self.test_diecase_charset, seq=15,
                          text='Test if diecase contains required characters'),
                   option(key='f', value=_import, seq=30,
                          text='Import layout from file'),
                   option(key='x', value=_export, seq=31,
                          text='Export layout to file'),
                   option(key='r', value=self.resize_layout, seq=89,
                          text='Change the diecase layout size',
                          desc=('Current: {} x {}'
                                .format(self.diecase.rows,
                                        self.diecase.columns))),
                   option(key='n', value=_clear_layout,
                          text='Clear the diecase layout', seq=90),
                   option(key='ctrl_s', value=_save, seq=91,
                          text='Store the diecase in database',
                          cond=self.diecase.diecase_id),
                   option(key='ctrl_n', value=_save_as, seq=92,
                          text='Store the diecase under a different name',
                          cond=is_stored),
                   option(key='delete', value=_delete, seq=93, cond=is_stored,
                          text='Delete diecase from database'),
                   option(key='F2', value=_change_diecase, seq=94,
                          text='Change diecase', cond=diecases_present),
                   option(key='F3', seq=95, text='List typefaces',
                          value=UI.paused(views.list_typefaces)),
                   option(key='F4', seq=96, text='List unit arrangements',
                          value=UI.paused(views.list_unit_arrangements)),
                   option(key='F5', seq=97, text='List normal wedges',
                          value=UI.paused(views.list_wedges))]
            return ret

        UI.dynamic_menu(options, header=header)



# -*- coding: utf-8 -*-
"""Constants and basic data structure definitions"""

from collections import namedtuple

# define some namedtuple objects so that returning data is easier
Coordinates = namedtuple('Coordinates', 'column row')
MatrixSize = namedtuple('MatrixSize', 'horizontal vertical')
MatrixRecord = namedtuple('MatrixRecord', 'char styles code units')
WedgePositions = namedtuple('WedgePositions', 'pos_0075 pos_0005')
WedgeLimits = namedtuple('WedgeLimits', 'shrink stretch')
QueueItem = namedtuple('QueueItem', 'matrix units qty')

# Style definitions
Style = namedtuple('Style', 'name alternatives short codes ansi')
SD = namedtuple('StyleDefinitions',
                ('roman bold semibold italic altitalic bolditalic '
                 'smallcaps fraktur inferior superior '
                 'size1 size2 size3 size4 size5'))
STYLES = SD(roman=Style(name='roman', short='r',
                        alternatives='regular, antiqua',
                        codes=('rr', '00'), ansi=0),
            italic=Style(name='italic', short='i', alternatives='',
                         codes=('ii', '01'), ansi='3;31'),
            altitalic=Style(name='alternative italic', short='a', ansi='3;31',
                            codes=('ai', '08'), alternatives=''),
            semibold=Style(name='semi-bold', short='h', ansi='1;33',
                           codes=('sb', '09'), alternatives=''),
            bold=Style(name='bold', short='b', alternatives='',
                       codes=('bb', '03'), ansi='1;33'),
            bolditalic=Style(name='bold italic', short='q', ansi='1;31',
                             alternatives='',
                             codes=('bi', 'ib', '07')),
            smallcaps=Style(name='small caps', short='s', alternatives='',
                            codes=('sc', 'ss', '02'),
                            ansi='4;32'),
            inferior=Style(name='inferior', short='l',
                           alternatives='lower index, subscript',
                           codes=('ll', '05'), ansi=34),
            superior=Style(name='superior', short='u',
                           alternatives='upper index, superscript',
                           codes=('uu', '04'), ansi=35),
            fraktur=Style(name='Fraktur', short='f',
                          alternatives='Schwabacher, German',
                          codes=('ff', '06'), ansi=36),
            size1=Style(name='size 1', short='1', ansi=0,
                        alternatives='', codes=('s1',)),
            size2=Style(name='size 2', short='2', ansi=0,
                        alternatives='', codes=('s2',)),
            size3=Style(name='size 3', short='3', ansi=0,
                        alternatives='', codes=('s3',)),
            size4=Style(name='size 4', short='4', ansi=0,
                        alternatives='', codes=('s4',)),
            size5=Style(name='size 5', short='5', ansi=0,
                        alternatives='', codes=('s5',)))
# Style control commands for typesetting - generate automatically
STYLE_COMMANDS = {code: style for style in STYLES for code in style.codes}

# Text alignments
Alignment = namedtuple('Alignment', 'name alternatives codes')
AD = namedtuple('AlignmentDefinitions', 'left center right both')
ALIGNMENTS = AD(left=Alignment(name='left', codes=('cr',),
                               alternatives='flush left, rag right'),
                center=Alignment(name='center', codes=('cc',),
                                 alternatives='rag left and right'),
                right=Alignment(name='right', codes=('cl',),
                                alternatives='flush right, rag left'),
                both=Alignment(name='both', codes=('cf', 'cb'),
                               alternatives='justified, flat, flush both'))
ALIGN_COMMANDS = {code: algmt for algmt in ALIGNMENTS for code in algmt.codes}

# Default space positions
DEFAULT_LOW_SPACE_POSITIONS = (('G', 1), ('G', 2), ('G', 5), ('O', 15))
DEFAULT_HIGH_SPACE_POSITIONS = (('O', 16),)
# Names for low and high spaces, depending on their width
SPACE_NAMES = {'   ': 'low em quad', '  ': 'low en quad', ' ': 'low space',
               '___': 'high em quad', '__': 'high en quad', '_': 'high space'}
SPACE_SYMBOLS = {' ': '□', '_': '▣'}

# Measurement units for line length, character/space width etc.
TYPOGRAPHIC_UNITS = dict(pc=12.0, pt=1.0, Pp=12*0.166/0.1667, pp=0.166/0.1667,
                         cc=12*0.1776/0.1667, dd=0.1776/0.1667,
                         cf=12*0.1628/0.1667, ff=0.1628/0.1667,
                         cm=0.3937*72, mm=0.03937*72)
TYPOGRAPHIC_UNITS['in'] = TYPOGRAPHIC_UNITS['"'] = 72.0


# S5 wedge unit values; wedge name aliases
S5 = [5, 6, 7, 8, 9, 9, 9, 10, 10, 11, 12, 13, 14, 15, 18]


# Accented letters will use the same unit arrangement value as their
# non-accented counterparts.
ACCENTS = dict(a='ąäáãâạà', c='ćĉç', e='ęëȩẽéêèẹ', g='ǵĝ', h='ḧḩĥḥ',
               i='ïĩíîìị', j='j́ĵ', k='ķḱḳ', l='łļĺḷ', m='ḿṃ', n='ńņñǹṇ',
               o='óöõôòọ', p='þṕ', r='ŗŕṛ', s='śşŝṣ', t='ţẗṭ', u='üũúûùụǘǜ',
               v='ṽṿ', w='ẅẃŵẁẉ', x='ẍ', y='ÿỹýŷỳỵ', z='żźẑẓ')
ACCENTS.update({letter.upper(): ACCENTS[letter].upper() for letter in ACCENTS})
REVERSE_ACCENTS = {accented_char: unaccented_char
                   for unaccented_char, accents in ACCENTS.items()
                   for accented_char in accents}

# Typical ligatures:
LIGATURES = ['ae', 'oe', 'AE', 'OE', 'fi', 'fl', 'ff', 'fb', 'fh', 'fk',
             'fj', 'ffi', 'ffl', 'ij', 'IJ', 'st', 'ct']

# Language codes
LANGS = {'en': 'English', 'nl': 'Dutch', 'pl': 'Polish', 'de': 'German',
         'eo': 'Esperanto', 'tr': 'Turkish', 'it': 'Italian',
         'cz': 'Czech', 'fr': 'French', 'es': 'Spanish', 'pt': 'Portugese',
         'da': 'Danish', 'fi': 'Finnish', 'sv': 'Swedish', '#': 'numbers'}
